/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
     author: kekec 		E-mail: behaviour.brain.research@gmail.com
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          dwp version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var item_cnt = 10;    // The total number of items (stimuli) in the task

var condit = '11';  // Condition type : '1'.. '11'. Select the type of word the user must type in the answer field

var c_min = 3;  // Minimum required word length

var c_max = 10; // Maximum required word length

var is_ltr = true;   // Parsing (reading) direction: true => '1' = left-to-right or false => '2' = right-to-left 

var is_lower = '2'; // Letter type: '1' = Original, '2' = lowercase; '3' = UPPERCASE

var user_del = 'enter'; // Delimiter character between the words in the word entry multi-line text box. '\n' is not used here because it must be escaped 

var all_c = true;   // All letters of the alphabet are used: Yes = '1' = true or 'No' = '2' = false

var limen = 2;  // Minimum word length to be accepted as valid (threshold) => document.getElementById('threshold').value
 
var timer_enabled = true;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var timer = 600;     // Task time in seconds if clock = 1000ms (default). 

// var font_family = 'Arial, Helvetica, sans-serif';  // user answer fontFamily. only local scope in initLayout() 

// var font_size= 'medium';  // font size of the text. . only local scope in initLayout() 

var is_back = true;    // if (is_back==true) => user can go back to previous task/item. (true || false)

var prg_type = '1';   //  type of information in Progress Bar ('0' => Items, '1' => Time; '2' => Nothing) 

var demo = true;       // if (demo == true) => Show task solutions 

// var stat = '0'; // show the number of words in the dictionary that meet the set conditions. only local scope.

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global variables:

const clock = 1000;  // basic timer resolution in ms

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var startTime;        // time when task started (ex counter)

var item;   				// current item in task. range: [0..item_cnt-1]

var is_task = false;    // task in progress or not. It is true for all events between end of startTask() and theEnd() events

var dirty = false;  	// if  (dirty==true) => user changed any options in form 

var is_storage = false;		// local storage is enabled on user browser or not
  
//  global objects and arrays declarations..

var letters = { a:1,  b:1,  c:1,  d:1,  e:1,  f:1,  g:1,  h:1,  i:1,  j:1,  k:1,  l:1,  m:1,  n:1,  o:1,  p:1,  q:1,  r:1,  s:1,  t:1,  u:1,  v:1,  w:1,  x:1,  y:1,  z:1 };  // Alphabetic letters object. Default: English alphabet (Latin alphabet consisting of 26 letters)

var char_def = true;  // default Alphabetic letters array is used (true) or not (user loaded custom letters from txt file) 

var dic_def = true;  // default Dictionary is used (true) or not (user loaded custom Dictionary loaded from txt file) 

/* stim = array of objects. For each item in task [0..item_cnt-1] is this object:  
  txt = String displayed to used
  cond = Type of condition (window.condit)
  data = specific data for this condition type (eq. required chars begin/end, contains, required word length). see stimGen() 	
*/
var stim = [];   

var ans = [];    // user answers array (html textarea)

var freq = {};  // objects with 2 keys. used only if document.getElementById('stat').value == '1'. At the end of the task, show the number of words in the dictionary that meet the set conditions (requests). see: StimGen(), doStat(), theEnd()

var rndKey = function (obj) {
    let keys = Object.keys(obj);
    return keys[ keys.length * Math.random() << 0];
};

/*
var rndValue = function (obj) {
    let keys = Object.keys(obj);
    return obj[keys[ keys.length * Math.random() << 0]];
}; */



/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */



// application init [create/start Timer, check storage, show start panel..]
function init() 
{ 
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
	
	is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }	
		
	loadFont();
	
	setIns();
	
	document.getElementById('letters_cnt').textContent= Object.keys(letters).length;  // Number of letters of the alphabet (26 or loaded from txt file) 
	
	document.getElementById('words_cnt').textContent= Object.keys(words).length;  // Number of words loaded from the dictionary
	
	setChar();
	
	showStart();  
}

 
/* application timer handler function 
   https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date
	 https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/getTime
	 Reduced time precision: In Firefox, the privacy.reduceTimerPrecision preference is enabled by default and defaults to 20µs in Firefox 59; in 60 it will be 2ms. */
function appTimer()  
{   
	if (timer_enabled && is_task)   // else: nop, idle, do nothing 
	{
		let endTime = new Date();  
		let seconds = Math.round((endTime.getTime() - startTime.getTime()) / 1000);  // elapsed time in seconds
		if ( seconds > timer )  {     // task time is limited
			theEnd(); 
		}	
		else if (prg_type == '1')   {   // '1' => Time. display time on Progress Bar  
			let rest= timer - seconds;   // this is remaining time 
			let percent= Math.round(100*rest/timer);    
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('nav_items').textContent= stringTime(rest);	
		}
	}  
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 (probably redundant)
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  let hh = Math.floor(total/(60*60));         
  let div_min = total % (60 * 60);
  let mm = Math.floor(div_min / 60);       
  let div_sec = div_min % 60;
  let ss = Math.ceil(div_sec);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex= 3; // 'Arial, Helvetica, sans-serif'
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[3];
}	


// show start panel 
function showStart()  
{     
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('dwp-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('dwp-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // xxx
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
	document.getElementById('no_task').style.display= 'none';  // specific to this app
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
	refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		
	
	let e= document.getElementById('user_del'); //  Delimiter character between the words in the word entry multi-line text box
  if ( e.selectedIndex == -1 ) {
	  e.selectedIndex= 0; // '\n' or Newline	
	}
  let ind= e.selectedIndex;
	user_del= e.options[e.selectedIndex].getAttribute('data-value');  // update Delimiter character
		
	document.getElementById('ins_user_del').textContent= ind == 0 ? 'Enter ONLY ONE word in each line of the text box.' : 'Separate individual words within the text box with a ' + e.options[ind].text + ' ("' + e.options[e.selectedIndex].getAttribute('data-value') + '").' ; 
	
	document.getElementById('ins_dic_def').style.display= dic_def ? 'block' : 'none';
	
	document.getElementById('ins_limen').textContent= 'If you want your word to be valid, the minimum number of letters used must be: ' + limen;
	
	document.getElementById('ins_back').style.display= is_back ? 'none' : 'block';
	
	if ( document.getElementById('timer_enabled').checked )
  {   
    document.getElementById('ins_timer').setAttribute('style', 'display:block'); 		
		document.getElementById('ins_timer').textContent= 'The duration of this task is limited.';  // The duration of this task is always limited.
		if (prg_type == '0')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will show how many items are left until the end of the task.';
		}
		else if (prg_type == '1')  {
			document.getElementById('ins_timer').textContent+= ' The Progress bar will display the remaining time until the end of the task.';  
		}
		else  {
			document.getElementById('ins_timer').textContent+= ' But information about the remaining time will not be visible.';  
		}  
  }
	else  {
		document.getElementById('ins_timer').setAttribute('style', 'display:none');
	}	
		
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';		
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty = true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ', ind;               
	switch (my_id)
	{
		
		case 'item_cnt':
			if (e.validity.valid)  {
				msg.textContent= ''; 				
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
			} 
			break; 
				
		case 'condit':
		  ind= e.selectedIndex;
			document.getElementById('length_div').style.display =  ind == 0 || ind == 10  ? 'block' : 'none';
		  break;  
		
		case 'c_min':
		case 'c_max':
			checkRange();
			checkLimen();
			break; 
   	 
    case 'is_ltr': 		
		  switch ( e.value ) {
				case '1':
			    msg.textContent= 'Left-to-right text direction.';
					break;
			  case '2':
			    msg.textContent= 'Right-to-left text direction.';
					break;		       
			}	
			msg.textContent += ' This only applies to direction of reading for the typed words within the text box for entering words. The application will process all typed words in this direction.';			
			break; 
		 
		case 'is_lower':	// ok
			switch ( e.value )	{
				case '1':
					msg.textContent= 'The text box for entering words will display the letters exactly as the user enters them. The text renders as it is. ';
					break;
				case '2':
					msg.textContent= 'The text box for entering words will only display lowercase letters.';
					break;	
				case '3':
					msg.textContent= 'The text box for entering words will only display UPPERCASE letters.';
					break;		
      }     
			break; 			
			
		case 'user_del':	
			ind= e.selectedIndex;
			if ( ind == 0 )  {
				msg.textContent=  'The user only needs to enter one word in each line of the text box. This is probably the safest option ';
				msg.textContent+=  '(' + e.options[e.selectedIndex].getAttribute('title') + ')'; 
			}	
			else if ( ind > 0 ) {
				msg.textContent= 'Make sure that no word in the dictionary used contains the selected Delimiter: ' + e.options[ind].text; 
				msg.textContent+=  ' ("' + e.options[e.selectedIndex].getAttribute('data-value') + '")'; 
			}
		  break;		
		
		case 'all_c':	 
		  if (e.value == '1') {					
				msg.textContent= 'All letters will be used in the requests (pseudorandom).';
				document.getElementById('d_all_c').style.display= 'none';
			} 
      else  {
				msg.textContent= 'Only selected letters will be used in the requests (minimum 3).';
				document.getElementById('d_all_c').style.display= 'block';
			}				
			break;
			
    case 'threshold':
			if (e.validity.valid)  {
				msg.textContent= 'This is the minimum validity threshold. All typed words that are less than this threshold will be invalid (0 points), regardless of whether the word is in the dictionary or not.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			checkLimen();		
			break; 	

		case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes.';
        document.getElementById('p_timer').style.display= 'block';
      }
      else  {  
        msg.textContent= 'No. Timer is disabled.';
        document.getElementById('p_timer').style.display= 'none'; 
      }       
      break;
		
    case 'timer':
			if (e.validity.valid)  {
				msg.textContent= stringTime(e.value); 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 

		/* case 'font_family': => uses tryFont()
		  break; */
		
		case 'font_size':
			msg.style.fontSize = e.value;
			//document.getElementById('font_family_msg').style.fontSize = e.value;
		  break;
		
		case 'is_back':		
			msg.textContent= e.checked ? 'Yes.' : 'No.';      
			break; 
						
		case 'prg_type':                                                                                                                                            
      if (e.value == '0')  {
        msg.textContent= 'It will be displayed how many items are left to the end.'; // The number of the current item will be displayed, as well as the total number of items.
      }
      else if (e.value == '1')  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break;    

		case 'demo': 
			msg.textContent= e.checked ? 'On. If you click the Check button at the bottom of the screen, you will receive feedback on the validity of all typed words.' : 'Off';  
			break; 
			
		case 'c_del_l':		// Load the letters of the alphabet. Delimiter character
			document.getElementById('s_del_l').style.display= e.checked ? 'none' : 'inline';     
			break; 
		
    case 'c_del_w':	// 	Load the dictionary. Delimiter character
			document.getElementById('s_del_w').style.display= e.checked ? 'none' : 'inline';     
			break;

    case 'stat':
		  let itm= document.getElementById('item_cnt');
			itm.max= e.selectedIndex == 0 ? 60 : 30;
			document.getElementById('item_cnt_msg').textContent= itm.validity.valid ? '' : s + itm.min + ".." + itm.max + s2 + itm.step; 
			break;	

		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;	
 
	}   // switch
	
  refMsg();   
}


// check c_min <= c_max. display warnings 
function checkRange()  
{
  let c_min = document.getElementById('c_min');
  let c_max = document.getElementById('c_max');
  let c_min_msg = document.getElementById('c_min_msg');
  let c_max_msg = document.getElementById('c_max_msg');  
  let min_v = parseInt(c_min.value,10), max_v = parseInt(c_max.value,10); // current values for min and max  
  if (c_min.validity.valid && c_max.validity.valid)   
  {                      
    if  ( min_v > max_v )  
    {  
      let tmp = 'Minimum required word length must not be higher than the Maximum required word length.';
			c_min_msg.textContent= tmp;  
      c_max_msg.textContent= tmp; 
      c_min_msg.setAttribute('class', 'error');	
      c_max_msg.setAttribute('class', 'error');						
    } 
		else if  ( min_v == max_v )  { 
		  let tmp= 'If these 2 values are equal, the same condition will be repeated over and over again.';
		  c_min_msg.textContent= tmp;  
      c_max_msg.textContent= tmp;  
			c_min_msg.removeAttribute('class');
		  c_max_msg.removeAttribute('class');
		}
    else   
    {                                                                                               
      c_min_msg.textContent= '';  
      c_max_msg.textContent= ''; 
     	c_min_msg.removeAttribute('class');
		  c_max_msg.removeAttribute('class');	
    }
  } 
  else  
  {
		let s = 'Values within this range are allowed: ', s2= ' with step= ';    
    c_min_msg.textContent= s + c_min.min + ".." + c_min.max + s2 + c_min.step;
		c_max_msg.textContent= s + c_max.min + ".." + c_max.max + s2 + c_max.step;
		c_min_msg.setAttribute('class', 'error');	
    c_max_msg.setAttribute('class', 'error');	
  }    
}


// check Limen/Treshold. display warnings => The selected threshold value must be less than or equal to Minimum required word length
function checkLimen()  
{
	let cond = document.getElementById('condit').value; 
	if ( cond == '1' || cond == '11' ) {
		let c_min = parseInt(document.getElementById('c_min').value, 10), c_max = parseInt(document.getElementById('c_max').value, 10); 
		let limen_max= Math.min(c_min, c_max), msg = document.getElementById('threshold_msg')
		if ( parseInt(document.getElementById('threshold').value, 10) > limen_max ) { 	
			msg.textContent= 'The selected threshold value must be less than or equal to ' + limen_max + ' (Minimum required word length).';
			msg.setAttribute('class', 'error');
		}	
		else  {
			msg.textContent= 'This is the minimum validity threshold. All typed words that are less than this threshold will be invalid (0 points), regardless of whether the word is in the dictionary or not.'; 
     	msg.removeAttribute('class');
		}
	}					
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity most of options 
// for other options see setPar()
function optionsValid()  
{ 
  let res = document.getElementById('item_cnt').validity.valid 
				 &&	document.getElementById('c_min').validity.valid 
				 && document.getElementById('c_max').validity.valid
         && document.getElementById('threshold').validity.valid				 
				 && document.getElementById('timer').validity.valid;
	
	if ( parseInt(document.getElementById('c_min').value,10) > parseInt(document.getElementById('c_max').value, 10) )  {  
		res = false;
	}		
	let condit = document.getElementById('condit').value; 
	if ( condit == '1' || condit == '11' ) {
		let c_min = parseInt(document.getElementById('c_min').value, 10), c_max = parseInt(document.getElementById('c_max').value, 10); 
		if ( limen > Math.min(c_min, c_max) ) {  
			res= false;
		}	
	}
  return res;              
}


// start task (research) 
function startTask(e)   
{      			 	
  if ( setPar() == true )  {		
				
		initLayout(); 

		stimGen();
		
		item = 0;
		
		startTime = new Date();
		
		is_task= true;   // if ( timer_enabled ) finally "activate" timer 
		
		drawStim();    // draw first item (stimulus)
	}
	// else do nothing. wait for user to change settings
}


// init DOM/html elements (positions,visibility,..)
function initLayout()
{	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
	document.getElementById('task_section').style.display= 'block'; 	  	
	
 	document.getElementById('progress').style.display= 'block';   //document.getElementById('progress').setAttribute('style', 'width:100%');	 
	if ( document.getElementById('prg_type').value == '2' || ( document.getElementById('prg_type').value == '1' && timer_enabled == false ) ) {
		document.getElementById('progress_holder').style.display= 'none'; 
	}	
	else  
	{
		document.getElementById('progress_holder').style.display= 'block';     
		//document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value == '0') {
			document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     //  items / remaining items
		}
		else if (document.getElementById('prg_type').value == '1')  {
			document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}					
	document.getElementById('nav_items').textContent='';  // always reset  
  //document.getElementById('feedback').textContent='';  // in drawStim()
	
	document.getElementById('demo_div').style.display= demo ? 'block' : 'none';   // Check button visible or not
	
	// object.style.textTransform = "none|capitalize|uppercase|lowercase|initial|inherit"
	if ( is_lower == '2' )  {
		document.getElementById('answer').style.textTransform = 'lowercase';
	}
	else if ( is_lower == '3' )  {
		document.getElementById('answer').style.textTransform = 'uppercase';
	}
	else  {
		document.getElementById('answer').style.textTransform = 'none'; // No characters are transformed. This is default
  }	
		
	document.getElementById('note').textContent= document.getElementById('ins_user_del').textContent;
	
	let font_size= document.getElementById('font_size').value;
	document.getElementById('question').style.fontSize = font_size;
	document.getElementById('answer').style.fontSize =  font_size;

  let font_family= document.getElementById('font_family').value;	
	document.getElementById('question').style.fontFamily= font_family;
  document.getElementById('answer').style.fontFamily= font_family;
	document.getElementById('note').style.fontFamily= font_family;
	document.getElementById('feedback').style.fontFamily= font_family;	
	
}  


/* set internal application parameters (global vars)
	return true if task can be started (or false if not)
  set content of id="no_task" validator message  */
function setPar()  
{  
	let ok = true;  // function result
	
	if ( document.getElementById('stat').value == '1' && item_cnt > 30 )  {
		item_cnt= 30;
		document.getElementById('item_cnt').value= 30; 
	}
	
	if ( document.getElementById('condit').selectedIndex < 0 ) {
		document.getElementById('condit').selectedIndex= 0;
		condit= document.getElementById('condit').value;   // condit= '1'  => each word must be of a certain length
	}
	
	if ( c_min > c_max ) {  // this should be impossible but because of danger of indefinite loop in stimGen()
    let v = Math.min(c_min, c_max);
		c_min= v;
		c_max= v;
		document.getElementById('c_min').value= v;
		document.getElementById('c_max').value= v;
	}	

  document.getElementById('no_task').innerHTML= ''; 
	/*if ( c_min < 3 || c_max < 3 ) {
		document.getElementById('no_task').innerHTML+= 'Minimum and Maximum required word length must be greater than 2.<br>';
		ok= false;
	}*/

  let condit = document.getElementById('condit').value; 
	if ( condit == '1' || condit == '11' )  // '1' => "each word must be of a certain length" or '11' => "all possible conditions - mixed"
	{	
		if ( limen > Math.min(c_min, c_max) ) {  
			document.getElementById('no_task').innerHTML = 'Minimum word length to be accepted as valid (threshold) is too big. The maximum allowed value is = ' + Math.min(c_min, c_max) + '. Currently set to = ' + limen + '.<br>';
			ok= false;
		}	
  }
	
	// this is probably redundant check (this section should never happen)
	let let_cnt = Object.keys(letters).length;  // Check number of letters of the alphabet (26 or loaded from txt file) 
	if ( let_cnt < 3 ) {
		document.getElementById('no_task').innerHTML+= 'The number of loaded letters of the alphabet is less than 3. Check the uploaded file or reload the current document, ie press the reload button in your browser to reload the default settings.<br>';  	
		ok = false;
	}
  else if ( let_cnt > 247 ) { 
		document.getElementById('no_task').innerHTML+= 'The number of letters of the alphabet used must be less than 248 (Tamil=247).<br>';
		ok = false;		
	} 
	/*if ( c_min > let_cnt || c_max > let_cnt) {
		document.getElementById('no_task').innerHTML+= 'Minimum and Maximum required word length must be less than the number of letters of the alphabet.<br>';
		ok = false;
	}*/	
		
  if ( !all_c )  
	{
		let col = document.getElementById('char_req').getElementsByClassName('char-req'), len = col.length, cnt= 0;  // live HTMLCollection: class name for required chars in parent containter 
		for (let j= 0; j < len; j++)  {
			let e = col[j];  // checkbox element (alphabet chars)
      if ( e.checked === true ) {
        cnt++;
			}
		}
		if ( cnt < 3 ) {
			document.getElementById('no_task').innerHTML+= 'If you do not want all letters to be used in the requests, the minimum number of checked (selected) letters must be 3.<br>';
			ok= false;
		}   
	}
	
	let wor_cnt= Object.keys(words).length;   // Number of words loaded from the dictionary	
	if ( wor_cnt == 0 ) {
		document.getElementById('no_task').innerHTML+= 'The number of words loaded from the dictionary is zero. Check the uploaded file or reload the current document, ie press the reload button in your browser to reload the default settings.<br>';
		ok = false;
	}
	/*else if ( item_cnt > wor_cnt ) {
		document.getElementById('no_task').innerHTML+= 'The number of items is greater than the total number of words loaded from the dictionary.<br>';
		ok= false;		
	}*/
	
	if ( ok ) {  // init task() => startTask()
		document.getElementById('no_task').style.display= 'none';
		return true;
	}
	else  {
		document.getElementById('no_task').innerHTML += ' Unable to start task.'
		document.getElementById('no_task').style.display= 'block';
		return false;
  }	
}


/* generate all stimuli for each item in range 0..item_cnt-1	
	Conditions:
	case '1':  // each word must be of a certain length => Type as many words as possible that contain exactly N letters (characters)
	case '2':  // each word must begin with a certain letter => Type as many words as possible starting with the letter X
	case '3':  // each word must end with a certain letter => Type as many words as possible ending with the letter X
	case '4':  // each word must contain a specific letter => Type as many words as possible that contain the letter X
	case '5':  // each word must not contain a specific letter => Type as many words as possible that don't contain the letter 
	case '6':  // each word must begin with one specific letter and end with another specific letter => Type as many words as possible that begin with the letter X and end with the letter Y	
	case '7':  // each word must contain 2 specific letters => Type as many words as possible that contain these 2 letters: X, Y
	case '8':  // each word must not contain 2 specific letters => Type as many words as possible that don't contain these 2 letters: X, Y
	case '9':  // each word must contain 3 specific letters => Type as many words as possible that contain these 3 letters: X, Y, Z
	case '10': // each word must not contain 3 specific letters => Type as many words as possible that don't contain these 3 letters: X, Y, Z
	case '11': // all possible conditions - mixed 
	*/
function stimGen()	
{		
	stim = [];	      // reset global window stimulus array 
	ans = [];         // reset global window user answer array 
	freq = {};        // reset global frequencies object
	let msg = { 
			//1: "Type as many words as possible that contain exactly N letters (characters)", 
		2: "Type as many words as possible starting with the letter ",
		3: "Type as many words as possible ending with the letter ",
		4: "Type as many words as possible that contain the letter ",
		5: "Type as many words as possible that don't contain the letter ",
			//6: "Type as many words as possible that begin with the letter XX and end with the letter YY",
		7: "Type as many words as possible that contain these 2 letters: ",
		8: "Type as many words as possible that don't contain these 2 letters: ",
		9: "Type as many words as possible that contain these 3 letters: ",
		10: "Type as many words as possible that don't contain these 3 letters: "
	};   	
	
	let u_letter= {}; // u_letters => used letters in this task => At the beginning of the task, a new collection of selected characters is created each time 
	if ( all_c )  {
		u_letter = letters;
	}
	else  
	{ // pick-up user selected letters and set new let_cnt (letters count)
		let col = document.getElementById('char_req').getElementsByClassName('char-req'), key;  //let col = document.getElementById('char_req').querySelectorAll('input[type=checkbox]').length;
		for (let i = 0, len = col.length; i < len; i++)  
		{
			if ( col[i].checked ) {
				key = col[i].value;  // key => check.value 
				u_letter[key] = 1;	 	
			}
		}
	}
	let let_cnt = Object.keys(u_letter).length;  //  update number of letters of the alphabet used it this task

	/* 
		 condition = selected condition (string converted to integer)
		 len_range =  total allowed range for word length		
	   con = previously selected condition type (only for mixed Condition type)
     size = previously selected word length	
	   chr = previously selected letters/chars => stim[i].data
		 rnd = for all pseudorandom values  
	*/	
	let condition = document.getElementById('condit').selectedIndex + 1, len_range = c_max - c_min + 1, con = {}, size = {}, chr = {}, rnd;  
  let stat = document.getElementById('stat').value == '1' ? true: false;  // At the end of the task, show the number of words in the dictionary that meet the set conditions (requests)

	for (let i= 0; i< item_cnt; i++)  
	{  
		ans[i]= [];
		if ( condit == '11' )   // mixed Condition type. change condition type for each item 
		{		
			if ( Object.keys(con).length  >= 10  )  {
				con = {};  // reset previous selected condition type (maximum reached) 
			}
			do {
				condition  = getRndInt(1,10);  
			} while ( con[condition] == 1 );	 
			con[condition]= 1;		
		}				
		switch (condition)
		{
			case 1:  // each word must be of a certain length => Type as many words as possible that contain exactly N letters (characters)
				if ( Object.keys(size).length  >= len_range  )  {
					size = {};  // reset previous selected word length (maximum reached) 
				}
				do {
				  rnd = getRndInt(c_min,c_max);
				} while ( size[rnd] == 1 );	 
				size[rnd]= 1;

				stim[i]= {
					txt : 'Type as many words as possible that contain exactly ' + rnd + ' letters (characters)',  // txt = String displayed to user
					cond : condition,    // cond = Type of condition (window.condit)
					data : rnd           // data = specific data for this condition type. in this case required word length
				};  		
				break;
				/*
				case 2:  // each word must begin with a certain letter => Type as many words as possible starting with the letter X
				case 3:  // each word must end with a certain letter => Type as many words as possible ending with the letter X
				case 4:  // each word must contain a specific letter => Type as many words as possible that contain the letter X
				case 5:  // each word must not contain a specific letter => Type as many words as possible that don't contain the letter 
				case 6:  // each word must begin with one specific letter and end with another specific letter => Type as many words as possible that begin with the letter X and end with the letter Y	
				case 7:  // each word must contain 2 specific letters => Type as many words as possible that contain these 2 letters: X, Y
				case 8:  // each word must not contain 2 specific letters => Type as many words as possible that don't contain these 2 letters: X, Y
				case 9:  // each word must contain 3 specific letters => Type as many words as possible that contain these 3 letters: X, Y, Z
				case 10: // each word must not contain 3 specific letters => Type as many words as possible that don't contain these 3 letters: X, Y, Z
				*/
			  default:
					/*  chr = previous selected letters/chars => stim[i].data
					    cnt= number or random characters generated (for conditions), 
							chars = selected chars for this item and condition
							tmp = fynal text message for user  */
					let cnt, chars= '', tmp; 
					if ( condition <= 5 ) {   // 2,3,4,5
						cnt= 1;  // generate only 1 random char
					}
					else if ( condition <= 8 )  {  // 6,7,8
						cnt= 2; // generate 2 random chars
					}
          else  {   // 9, 10
            cnt= 3;  // generate 3 random chars
					}			
          if ( Object.keys(chr).length + cnt >= let_cnt ) { 				
						chr = {};  // reset previous selected letters/chars
					} 
					for (let k = 1; k <= cnt; k++)  { 		  
						do  { 
							rnd = rndKey(u_letter);  
						} while ( chr[rnd] == 1 );	 
						chr[rnd]= 1;
						chars+= rnd;   				
					}					
					if ( condition == 6 )  {  // set tmp = fynal text message for user (id="question")
						tmp= 'Type as many words as possible that begin with the letter ' + chars.charAt(0).toLocaleUpperCase() + ' and end with the letter ' + chars.charAt(1).toLocaleUpperCase();
					}
					else  {
						tmp= msg[condition] + chars.charAt(0).toLocaleUpperCase();
						for (let k = 1; k < cnt; k++)  {
              tmp+= ', ' + chars.charAt(k).toLocaleUpperCase(); 
            }							
					}
					stim[i]= {
						txt : tmp,    // txt = String displayed to user
						cond : condition,   // cond = Type of condition 
						data : chars 		// data =  target character(s)
					};    
					break;
		}	// switch (condition)
    
		if ( stat ) {
			if ( typeof ( freq[ stim[i].cond ] ) === 'undefined' ) {
				freq[ stim[i].cond ] = {};
			}
			freq[ stim[i].cond ][ stim[i].data ] = 0;
		}
	}	 // for (let i= 0; i< item_cnt; i++)  	

}


// return random integer in range [min..max]
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1) ) + min;
}

 
// draw (repaint) current stimulus           
function drawStim() 
{    		 
  document.getElementById('question').textContent= stim[item].txt;

  let e = document.getElementById('answer');	
	if ( user_del == 'enter')   {  
		e.value= ans[item].join('\n');
	}
	else if ( user_del == ' ')   { 
		e.value= ans[item].join(' ');
	} 
	else {
		e.value= ans[item].join(user_del + ' ');
	}

	updateFooter();	
	document.getElementById('feedback').textContent= '';  // always reset
	
	document.getElementById('answer').focus();
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
	document.getElementById('font_size_msg').style.fontFamily= e.value;
}	


// Changes application settings that can be changed directly by the user within the options form
// These variables below can be directly modified by the user on the options form
function setVarPar()  
{ 
  
	if (document.getElementById('item_cnt').validity.valid)  {  
	  item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
  }
	
	condit= document.getElementById('condit').value; 
	
	if (document.getElementById('c_min').validity.valid && document.getElementById('c_max').validity.valid 
    && (parseInt(document.getElementById('c_min').value,10) <= parseInt(document.getElementById('c_max').value, 10) ) )   {  
      c_min= parseInt(document.getElementById('c_min').value, 10); 
      c_max= parseInt(document.getElementById('c_max').value, 10); 
  }	
	
  is_ltr= document.getElementById('is_ltr').value == '1' ? true : false;   // '1' => ltr, '2' => rtl
	
	is_lower= document.getElementById('is_lower').value; 
	
  // user_del => see: setIns()
	
	all_c= document.getElementById('all_c').value == '1' ? true : false;   
		
	if ( document.getElementById('threshold').validity.valid ) {
		limen= parseInt(document.getElementById('threshold').value, 10);   // xx
	}	
			
  timer_enabled= document.getElementById('timer_enabled').checked ? true : false;  // timer is enabled or disabled	
	
	if (timer_enabled && document.getElementById('timer').validity.valid) {
		timer= Math.round(parseInt(document.getElementById('timer').value, 10));  // task time (s)
  }	
		
	is_back= document.getElementById('is_back').checked ? true : false; 
	
	prg_type= document.getElementById('prg_type').value;  
		
  demo= document.getElementById('demo').checked ? true : false;	
	
  store= (is_storage && document.getElementById('store').value === '1') ? true : false;

}


// goto previous item (stimulus)
function priorItem()  
{      
  if (item > 0)  {   
		saveAnswer();  
		item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{  
  saveAnswer();  		
	item++;					
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// store user answer in current item (from textarea html element)
function saveAnswer()
{
	ans[item]= [];
	let lines = document.getElementById('answer').value.split('\n');
	for (let i = 0, len= lines.length; i < len; i++)  
	{
	  let txt = lines[i].trim();  // text in i-line
		if ( txt != '' )  
		{
      if ( user_del == 'enter' )  {  
				ans[item].push(txt);   // whole trimmed line
			}
			else  
			{ // first split by user delimiter and then trim()
				let arr = txt.split(user_del), len= arr.length; 
				for (let j= 0; j < len; j++) {
					arr[j]= arr[j].trim();
					if ( arr[j] != '' )  {
						ans[item].push( arr[j] );
					}
				}
			}
    }	
	}
}


// update footer when show each stimulus in task. update buttons and update progress bar
function updateFooter()
{  
	document.getElementById('prior').style.display= item > 0 && is_back ? 'inline': 'none';  // set footer navigation
	let btn= document.getElementById('next');     
	if (item == item_cnt-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update 'Next item' image on 'right arrow' button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Go forward';
	}						
  if (prg_type == '0' )  {    // update Progress bar => show current item count and remaining item count on Progress Bar                                                                                                                                                       
		let percent = Math.round ( 100 * (item_cnt - item - 1) / item_cnt );      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent=  (item + 1) + ' / ' + (item_cnt - item - 1);  
	}
}


// set default or user-loaded (required) letters
function setChar()
{	
	let cont= document.getElementById('char_req'), cnt = 1;
	cont.innerHTML= '';
	for (let key in letters)  {
		let check = document.createElement('input');
    check.type = 'checkbox';
		check.setAttribute('checked', true);
		check.id = 'creq' + cnt;		
		//check.setAttribute('data-value', key);
		
    check.setAttribute('class', 'char-req');
		check.value = key;
    check.title= key;
 
    let lab = document.createElement('label');
    lab.htmlFor = check.id;
    lab.title= key;
		lab.appendChild(document.createTextNode(key));
    
    cont.appendChild(check);
    cont.appendChild(lab);
		
		//let br = document.createElement('br');
    //cont.appendChild(br);
		
		cont.innerHTML += '&Tab;';  // or: '&#9;' or: '&#x9;' or: '&Tab;'	
    cnt++;		
	}
}


// https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsText
// https://stackoverflow.com/questions/23331546/how-to-use-javascript-to-read-local-text-file-and-read-line-by-line
// Loads alphabetic letters from a text file
function loadLetters() 
{
  let file =  document.getElementById('file_let').files[0];   // document.querySelector('#file_let').files[0];
  let delimit = document.getElementById('c_del_l').checked ? '\n' : document.getElementById('delim_l').value;
	let reader = new FileReader(); 
	reader.addEventListener('load', function () 
	{	   
		let chars = reader.result.split(delimit);  // xxx new array (reader.result == a complete text file) 
		let len = chars.length;
		if ( len < 1 ) {
			alert('The number of alphabet letters loaded is zero. Make sure you are using the correct delimiter for the selected file. Check the uploaded file. Currently selected Delimiter character = ' + delimit);
		}
		else if ( len < 12 || len > 247) {
			alert('The number of letters of the alphabet used must be in this range: 12(Rotokas)..247(Tamil). Check the uploaded file. Currently selected Delimiter character = ' + delimit); 
		}
		else  
		{
			char_def= false;
			letters= {};
			let emp= 0, dup= 0;
			for (let i= 0; i< len; i++)  
			{
				let key= chars[i].trim();			
				if ( chars[i] ==  '' ) {
					emp++;
				}
				else if ( letters[key] || letters[key.toLocaleLowerCase()] || letters[key.toLocaleUpperCase()] )  {
					dup++;
				}
				else {
					letters[key] = 1;
				}
			}
			let let_cnt= Object.keys(letters).length;
			document.getElementById('letters_cnt').textContent= let_cnt + ' (custom Alphabet)';  // Number of letters of the alphabet (26 or loaded from txt file) 
			setChar();	
      alert('File upload complete: ' + file.name + ' \n Total number of loaded letters of the alphabet (characters) = ' + let_cnt + ' \n Number of blank (empty) characters skipped = '  + emp + ' \n Number of letters (characters) repeated = ' + dup + ' \n Total number of non-loaded characters = ' + (emp + dup) ); 
			
		}		
  }, false);
  /*reader.addEventListener('error', function() {
    alert('Error occurred reading file: ' + file.name);
  });*/

  if (file) {
    reader.readAsText(file);
  }
}


// https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsText
// text html: https://www.howtoexcel.org/split-text-by-delimiter/
// Loads the user's dictionary from a text file
function loadDictionary() 
{
  let file = document.getElementById('file_dic').files[0];
  let delimit = document.getElementById('c_del_w').checked ? '\n' : document.getElementById('delim_w').value;
	let reader = new FileReader(), stimuli, key, emp= 0, dup= 0;
	reader.addEventListener('load', function () 
	{	
	  stimuli = reader.result.split(delimit);
		let tot = stimuli.length;
		if ( tot < 1 ) {
			alert('The number of words loaded is zero. Make sure you are using the correct delimiter for the selected file. Check the uploaded file. Currently selected Delimiter character = ' + delimit);
		}
		else  
		{
			words= {}; 
			for (let i = 0; i < tot; i++)	
			{			
				key = stimuli[i].replace(/\r\n|\n|\r/g, '').trim();  // stimuli[i].toString().replace(/\r\n|\n|\r/g, '').trim();  
				if ( key == '' ) {
					emp++;
				}
				else if ( words[key]  )  { 
					dup++;  // escape duplicates
				}
				else  {
					words[key] = 1;
				}
			}		
			let wor_cnt= Object.keys(words).length;   // Number of words loaded from the dictionary
			document.getElementById('words_cnt').textContent= wor_cnt + ' (custom Dictionary)';  // Update number of words loaded from the dictionary
			dic_def= false;   // default Dictionary is not used => user loaded custom Dictionary loaded from txt file 
			
			alert('File upload complete: ' + file.name + ' \n Total number of successfully loaded words = ' + wor_cnt + ' \n Number of empty (blank) words = '  + emp + ' \n Number of words repeated = ' + dup + ' \n Total number of non-loaded words = ' + (emp + dup) ); 
		}
	}, false);
  /*reader.addEventListener('error', function() {
    alert('Error occurred reading file: ' + file.name);
  });*/

  if (file) {
    reader.readAsText(file);
  }
}


// https://www.freecodecamp.org/news/how-to-reverse-a-string-in-javascript-in-3-different-ways-75e4763c68cb/
function reverseString(str) {
   return str.split("").reverse().join("");
}


// Check which typed words are correct (found in the dictionary used). only when demo == true
function checkWords()
{
	saveAnswer();   
	// for display trimmed() and auto-correct values call drawStim() here
	let res = getScore(item), points = res.length;
	let e = document.getElementById('feedback'); 
	if ( points == 0 )  { 
		e.textContent= "Number of points: 0. You didn't enter any valid words.";		// "You didn't enter any valid words.";
	}
	else  {
		e.textContent=  'Number of points: ' + points + '. Valid words = ' + res;
	}
	e.textContent +=  ' Minimum validity threshold = ' + limen + '.';
}	


// At the end of the task, show the number of words in the dictionary that meet the set conditions (requests)
function doStat() 
{
	for (let word in words) 
	{	// iterate through each word within the dictionary
		for (let cond in freq)  // 1. key in freq object is condition type
		{ // iterate through each type of condition
			for (let data in freq[cond])   // 2. key in freq object is data (1-3 chars or word length)
			{ // iterate through each data set within each condition type (the same condition type can be applied but eg with different chars)  
				if  ( isMatch(word, cond, data) ) {     //  word is not reversed in dictionary
					freq[cond][data]++;
				}
			}
		}	
	}	
}


/* check that the word meets the conditions. return true if word is ok. return false if word is not match selection criteria
  parameters: 
	word = word to test, cond = condition type, data = data specific for this condition type
	stim[i].cond = 1  Type as many words as possible that contain exactly N letters (characters) 
	stim[i].cond = 2. Type as many words as possible starting with the letter X
	stim[i].cond = 3. Type as many words as possible ending with the letter X
	stim[i].cond = 4. Type as many words as possible that contain the letter X
	stim[i].cond = 5. Type as many words as possible that don't contain the letter X
	stim[i].cond = 6. Type as many words as possible that begin with the letter X and end with the letter Y	
	stim[i].cond = 7. Type as many words as possible that contain these 2 letters: X, Y
	stim[i].cond = 8. Type as many words as possible that don't contain these 2 letters: X, Y
	stim[i].cond = 9. Type as many words as possible that contain these 3 letters: X, Y, Z
	stim[i].cond = 10. Type as many words as possible that don't contain these 3 letters: X, Y, Z	 	
  https://stackoverflow.com/questions/31068060/why-is-accessing-a-variable-using-window-variable-slower
  https://www.measurethat.net/Benchmarks/Show/4797/1/js-regex-vs-startswith-vs-indexof
  https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/indexOf 
  Searching for an empty search string produces strange results. With no second argument, or with a second argument whose value is less than the calling string's length, the return value is the same as the value of the second argument */	
function isMatch(word, cond, data)
{
		if ( word.length < limen) {
			return false;
		}
		let match, c;		
		switch ( cond ) 			// condition for this i-item
		{
			case '1':  // each word must be of a certain length. Type as many words as possible that contain exactly N letters (characters)
			  match = ( word.length == data )  ? true : false;
				break;			

			case '2': // each word must begin with a certain letter. Type as many words as possible starting with the letter X
			  c = word.charAt(0);  
				match=  ( c == data.toLocaleLowerCase() || c == data.toLocaleUpperCase() ) ? true: false;
			  break;

			case '3': // each word must end with a certain letter. Type as many words as possible ending with the letter X
				c= word.slice(-1);
				match= ( c == data.toLocaleLowerCase() || c == data.toLocaleUpperCase() ) ? true: false;
				break;

      case '4':  // each word must contain a specific letter => Type as many words as possible that contain the letter X							
				match = ( word.indexOf(data.toLocaleLowerCase()) > - 1 || word.indexOf(data.toLocaleUpperCase()) > - 1 ) ? true : false;
					/* https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/RegExp/ignoreCase
					let rgx = new RegExp(data, 'i');
					match = rgx.test(word); */		
				break;

			case '5':   // each word must not contain a specific letter => Type as many words as possible that don't contain the letter X						
				match = ( word.indexOf(data.toLocaleLowerCase()) != - 1 || word.indexOf(data.toLocaleUpperCase()) != - 1 ) ? false : true;
			//match = ( word.indexOf(data.toLocaleLowerCase()) == - 1 && word.indexOf(data.toLocaleUpperCase()) == - 1 ) ? true : false;
				break;	
			
      case '6':  // each word must begin with one specific letter and end with another specific letter => Type as many words as possible that begin with the letter X and end with the letter Y	
			  c = word.charAt(0);  
				match=  ( c == data[0].toLocaleLowerCase() || c == data[0].toLocaleUpperCase() ) ? true: false;
				if ( !match ) {
				  return false;
				}
				c= word.slice(-1);
				match= ( c == data[1].toLocaleLowerCase() || c == data[1].toLocaleUpperCase() ) ? true: false; 
				break;
				
			case '7':  // each word must contain 2 specific letters => Type as many words as possible that contain these 2 letters: X, Y
			case '9':  // each word must contain 3 specific letters => Type as many words as possible that contain these 3 letters: X, Y, Z			
        for (let k = 0; k < data.length; k++)  {
					c = data[k];
					if ( word.indexOf(c.toLocaleLowerCase()) == - 1 && word.indexOf(c.toLocaleUpperCase()) == - 1 ) {
						return false;
					}	
				}
				match= true;
        break;				
				
			case '8':  // each word must not contain 2 specific letters => Type as many words as possible that don't contain these 2 letters: X, Y	
			case '10': // each word must not contain 3 specific letters => Type as many words as possible that don't contain these 3 letters: X, Y, Z
			  for (let k = 0; k < data.length; k++)  {
					c = data[k];
					if ( word.indexOf(c.toLocaleLowerCase()) != - 1 || word.indexOf(c.toLocaleUpperCase()) != - 1 ) {
						return false;
					}	
				}
				match= true;
				break;
								
			/* default:  
         return null; */    			
		}
	  return match;
}	


// return array of valid words for item i
function getScore(i) 
{  
  // res => array (list) of all valid words for this item i
	// sel = object of already selected word (duplicates entry must be ignored)
	let res= [], sel={};    
	for (let j= 0, size = ans[i].length; j < size; j++ )
	{	 // score each word within current item
    let word = is_ltr ? ans[i][j].trim() : reverseString(ans[i][j]).trim();
		if ( sel[word.toLocaleLowerCase()] != 1 )
		{
			sel[word.toLocaleLowerCase()]= 1;
			if ( isMatch ( word, stim[i].cond.toString(), stim[i].data ) ) 
			{		
				
				let txt= ans[i][j].trim(), txt_low= txt.toLocaleLowerCase(), txt_up= txt.toLocaleUpperCase(), r_txt, r_txt_low, r_txt_up, ok;
				
				if ( is_ltr ) {  // parse left-to-right
					ok =  ( words[txt] == 1 || words[txt_low] == 1 || words[txt_up] == 1 );
				}
				else  {	// parse right-to-left
					r_txt= word, r_txt_low= r_txt.toLocaleLowerCase(), r_txt_up= r_txt.toLocaleUpperCase();
					ok = ( words[r_txt] == 1 || words[r_txt_low] == 1 || words[r_txt_up] == 1 );
				}	
					
				if ( ok ) 
				{  //  is_lower => Letter type: '1' = Original, '2' = lowercase; '3' = UPPERCASE			   
					if ( is_lower == '2' )  {
						res.push(txt_low);
					}
					else if ( is_lower == '3' )  {
						res.push(txt_up);
					}
					else  {
						res.push(txt);
					}
				}	
			}
		}		
	}	
	return res;
}	


// end of task (research)
function theEnd()  
{    
	is_task = false;

	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
	document.getElementById('result_wrapper').style.display= 'block';
  let stat = document.getElementById('stat').value == '1' ? true: false;  // At the end of the task, show the number of words in the dictionary that meet the set conditions (requests)
	if ( stat ) {
		doStat();
		document.getElementById('total_note').style.display= 'block';
		document.getElementById('percent_note').style.display= 'block';
	}
	else  {
		document.getElementById('total_note').style.display= 'none';
		document.getElementById('percent_note').style.display= 'none';
	}
	let header=  stat ? [ 'Item', 'Request (Condition)', 'Answer', 'Valid', 'Points', 'Total', 'Percent' ] : [ 'Item', 'Request (Condition)', 'Answer', 'Valid', 'Points' ]; // 5 or 7 columns
	let size= stim.length,  score = 0,  arr = [], rows = [], res, p, s, per;		
	let wor_cnt= Object.keys(words).length;   // Number of words loaded from the dictionary	
	for (let i = 0; i < size; i++)  // for each item in task => for (let i = 0; i < item_cnt; i++)		
	{					
		res = getScore(i);  // all valid words in this i-item 
		p= res.length;    // p => number of points in this i-item	
		arr.push(p);
		score+= p;  // sum of points in whole task		
		for (let j= 0, len= ans[i].length; j < len; j++)  {   // use =>  object.style.textTransform and not call reverseString..
			if ( is_lower == '3' )  {
				ans[i][j]= ans[i][j].toLocaleUpperCase();
			}
			else if ( is_lower == '2' )  {   
				ans[i][j]= ans[i][j].toLocaleLowerCase();
			}
		}
		s= ans[i]; //.join(',');  // user answer for this i-item
		stim[i].txt= stim[i].txt.replace('Type as many words as possible', '..');
		if (stat)  {
			per = 100 * freq [ stim[i].cond ][ stim[i].data ] / wor_cnt;
      per=  per == 0 ?  '0 (No words)' : per.toFixed(8);				
		  rows[i]= [ parseInt(i+1,10), stim[i].txt, s, res, p, freq [ stim[i].cond ][ stim[i].data ], per ]; 
		} 
		else {
			rows[i]= [ parseInt(i+1,10), stim[i].txt, s, res, p ]; 
		}
	}   // for (let i = 0; i < item_cnt; i++)			
		
	if (stat)  {
	  rows.push (  [ 'Sum',  '', '', '', score, '', ''  ] );   // 7. columns Sum => 'Item', 'Request (Condition)', 'Answer', 'Valid', 'Points', 'Total', 'Percent' 	
  }
	else  {
		rows.push (  [ 'Sum',  '', '', '', score  ] );   // 5. Columns Sum => 'Item', 'Request (Condition)', 'Answer', 'Valid', 'Points'
	}
	createTableArray(document.getElementById('table1'), rows, header); 
  document.getElementById('s_limen').textContent= limen;
	
	// table2 => Descriptive statistics	
	const D = 3;  // number of decimals
	rows = [];  // reset table rows (for table2)
	let n = arr.length;  
	let m =   n > 0  ?  mean(arr).toFixed(D) : '';      
	let sd =  n > 1  ?  stDev(arr, m).toFixed(D) : '';   	 	
  rows.push( [ m, sd, n ] );  // last row
	
	header= ['Average', 'Standard deviation', 'N']; // header		
	createTableArray(document.getElementById('table2'), rows, header); 

  // update 'score' label => above "Repeat" button
	document.getElementById('score').textContent= 'The sum of all points. Score= ' + score;
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	
	try 
  {
    if (store) { 
			setStorage(score, m, sd);   			
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  
}


// return arithmetic mean (mean, average) of input array (arr)
// no validation => input array must be in valid number format	
function mean(arr)  
{	
	let len= arr.length, sum= 0;
	for (let i=0; i<len; i++) {
		sum+= arr[i];
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation (with Bessel's correction) => N-1
// no validation => input array must be in valid number format
// arr = array argument;  mean = average for input arr 
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
		sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	if (dd < 10) { dd='0'+dd; } 
	if (mm < 10) { mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 score = sum (number) of points in task
   mean = average (mean) for user points 
   sd = standard deviation for user points	 */   
function setStorage(score, mean, sd)  
{    
	let cond = document.getElementById('condit').value, char_min, char_max; 
	if ( cond == '1' || cond == '11' ) {
	  char_min= c_min;
		char_max= c_max;
	}
	else  {
		char_min= '';
		char_max= '';
	}
	let stat = document.getElementById('stat').selectedIndex + 1;
	let oldItems = JSON.parse(localStorage.getItem('dwp-v1.0')) || []; 
  let id= oldItems.length + 1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
			'score' : score,
			'mean'  : mean, 	
			'sd' : sd,						
			'items' : item_cnt,
			'condition' : condit,
			'word-min' : char_min,  
      'word-max' : char_max,			
			'dir' : is_ltr ? 1 : 2,	
			'letter-type' : is_lower,
			'delimiter': user_del,
			'all-letter' : all_c ? 1 : 2,      // var all_c = true;  All letters of the alphabet are used: Yes = '1' = true or 'No' = '2' = false         
			'threshold' : limen,
			'timer' : timer_enabled ? timer : '',	
			'is-back' : is_back ? 1 : 0,
			'demo' : demo ? 1 : 0,
      'def-dic' : dic_def  ? 1 : 0,
      'def-let' : char_def ? 1 : 0,
      'stat': stat			
  };

  oldItems.push(newItem);

  localStorage.setItem('dwp-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
          col.push(key);
        }
    }
  } 
  let table = document.createElement('table');
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement('th');      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= '';
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('dwp-v1.0')  ?  JSON.parse(localStorage.getItem('dwp-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('dwp-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('dwp-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// copied from https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('dwp-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 	
	let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  
			break;
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
